/*
Copyright 2024 The Kruise Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

package main

import (
	"context"
	"log"

	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
	"k8s.io/client-go/rest"

	kruiseclientset "github.com/openkruise/kruise/pkg/client/clientset/versioned"
)

func main() {
	config, err := rest.InClusterConfig()
	if err != nil {
		panic(err)
	}
	kc, err := kruiseclientset.NewForConfig(config)
	if err != nil {
		panic(err)
	}
	cloneSets, err := kc.AppsV1alpha1().CloneSets("").List(context.Background(), metav1.ListOptions{Limit: 1})
	if err != nil {
		panic(err)
	}
	if len(cloneSets.Items) > 0 || cloneSets.Continue != "" {
		log.Fatalln("there still exists some clonesets in the cluster")
	}
	statefulSets, err := kc.AppsV1alpha1().StatefulSets("").List(context.Background(), metav1.ListOptions{Limit: 1})
	if err != nil {
		panic(err)
	}
	if len(statefulSets.Items) > 0 || statefulSets.Continue != "" {
		log.Fatalln("there still exists some advanced statefulsets in the cluster")
	}
	statefulSetsBeta1, err := kc.AppsV1beta1().StatefulSets("").List(context.Background(), metav1.ListOptions{Limit: 1})
	if err != nil {
		panic(err)
	}
	if len(statefulSetsBeta1.Items) > 0 || statefulSetsBeta1.Continue != "" {
		log.Fatalln("there still exists some advanced statefulsets in the cluster")
	}
	daemonSets, err := kc.AppsV1alpha1().DaemonSets("").List(context.Background(), metav1.ListOptions{Limit: 1})
	if err != nil {
		panic(err)
	}
	if len(daemonSets.Items) > 0 || daemonSets.Continue != "" {
		log.Fatalln("there still exists some advanced daemonsets in the cluster")
	}
	log.Println("cluster is clean, ready to delete kruise")
}
